import { highlight } from "@code-hike/lighter";
import { isGroup, isWhitespace, } from "./types.js";
import { splitAnnotationsAndCode } from "./extract-annotations.js";
export async function tokenize(codeWithoutAnnotations, lang, annotations, config) {
    const { theme } = config;
    const { code: c, annotations: a } = await splitAnnotationsAndCode(codeWithoutAnnotations, lang, config);
    const { lines } = await highlight(c, lang, theme, {
        annotations: [...annotations, ...a],
        scopes: true,
    });
    const tokens = joinLines(lines);
    // split surrounding whitespace for each token
    const splitTokens = splitWhitespace(tokens);
    // join consecutive whitespace tokens
    const joinedTokens = joinWhitespace(splitTokens);
    return joinedTokens;
}
// group the Lines into one array
function joinLines(lines) {
    const joinedTokens = [];
    lines.forEach((lineOrGroup) => {
        if ("lines" in lineOrGroup) {
            joinedTokens.push([
                [lineOrGroup.annotationName, lineOrGroup.annotationQuery],
                joinLines(lineOrGroup.lines),
            ]);
        }
        else {
            const tokens = joinTokens(lineOrGroup.tokens);
            joinedTokens.push(...tokens);
            joinedTokens.push("\n");
        }
    });
    return joinedTokens;
}
function joinTokens(tokens) {
    return tokens.map((tokenOrGroup) => {
        if ("tokens" in tokenOrGroup) {
            return [
                [
                    tokenOrGroup.annotationName,
                    tokenOrGroup.annotationQuery,
                    { inline: true },
                ],
                joinTokens(tokenOrGroup.tokens),
            ];
        }
        else {
            const t = [tokenOrGroup.content];
            const { color, ...rest } = tokenOrGroup.style || {};
            t.push(color);
            if (Object.keys(rest).length) {
                t.push(rest);
            }
            return t;
        }
    });
}
function splitWhitespace(tokens) {
    const ejected = [];
    tokens.forEach((tokenOrGroup) => {
        if (isGroup(tokenOrGroup)) {
            ejected.push([tokenOrGroup[0], splitWhitespace(tokenOrGroup[1])]);
        }
        else if (isWhitespace(tokenOrGroup)) {
            ejected.push(tokenOrGroup);
        }
        else {
            const [before, content, after] = splitSurroundingWhitespace(tokenOrGroup[0]);
            if (before === null || before === void 0 ? void 0 : before.length) {
                ejected.push(before);
            }
            if (content.length) {
                const copy = [...tokenOrGroup];
                copy[0] = content;
                ejected.push(copy);
            }
            if (after === null || after === void 0 ? void 0 : after.length) {
                ejected.push(after);
            }
        }
    });
    return ejected;
}
function joinWhitespace(tokens) {
    const joinedTokens = [];
    tokens.forEach((tokenOrGroup) => {
        if (isGroup(tokenOrGroup)) {
            joinedTokens.push([tokenOrGroup[0], joinWhitespace(tokenOrGroup[1])]);
        }
        else if (isWhitespace(tokenOrGroup)) {
            let last = joinedTokens[joinedTokens.length - 1];
            if (last && isWhitespace(last)) {
                joinedTokens[joinedTokens.length - 1] += tokenOrGroup;
            }
            else if (tokenOrGroup !== "") {
                joinedTokens.push(tokenOrGroup);
            }
        }
        else if (tokenOrGroup[0].length > 0) {
            joinedTokens.push(tokenOrGroup);
        }
    });
    return joinedTokens;
}
// splits " \t foo bar \n" into [" \t ","foo bar"," \n"]
// "foo bar" -> ["","foo bar",""]
function splitSurroundingWhitespace(content) {
    const trimmed = content.trim();
    const before = content.slice(0, content.indexOf(trimmed));
    const after = content.slice(content.indexOf(trimmed) + trimmed.length);
    return [before, trimmed, after];
}
