import { extractAnnotations } from "@code-hike/lighter";
export async function splitAnnotationsAndCode(code, lang, config) {
    var _a;
    let extractors = [extractCommentAnnotations];
    if (config.lines) {
        extractors.unshift(extractLineAnnotations);
    }
    let annotations = [];
    let codeWithoutAnnotations = code;
    for (const extractor of extractors) {
        const { code: newCode, annotations: newAnnotations } = await extractor(codeWithoutAnnotations, lang, config);
        annotations = [...annotations, ...newAnnotations];
        codeWithoutAnnotations = newCode;
    }
    // import external code if needed and re-run annotations extraction
    const fromAnnotations = annotations.filter((a) => a.name === "from");
    if (fromAnnotations.length === 1) {
        const fromData = (_a = fromAnnotations[0].query) === null || _a === void 0 ? void 0 : _a.trim();
        const [codepath, range] = (fromData === null || fromData === void 0 ? void 0 : fromData.split(/\s+/)) || [];
        const externalFileContent = await readFile(codepath, config.mdxPath, range);
        const { code: newCode, annotations: newAnnotations } = await splitAnnotationsAndCode(externalFileContent, lang, config);
        annotations = [
            ...annotations.filter((a) => a.name !== "from" && a.name !== "Line"),
            ...newAnnotations,
        ];
        codeWithoutAnnotations = newCode;
    }
    return { code: codeWithoutAnnotations, annotations };
}
async function extractCommentAnnotations(code, lang, { annotationPrefix }) {
    const extractor = (comment) => {
        var _a;
        // const regex = /\s*(!?[\w-]+)?(\([^\)]*\)|\[[^\]]*\])?(.*)$/
        const regex = new RegExp(`\\s*(${annotationPrefix}?[\\w-]+)?(\\([^\\)]*\\)|\\[[^\\]]*\\])?(.*)$`);
        const match = comment.match(regex);
        if (!match) {
            return null;
        }
        const name = match[1];
        const rangeString = match[2];
        const query = (_a = match[3]) === null || _a === void 0 ? void 0 : _a.trim();
        if (!name || !name.startsWith(annotationPrefix)) {
            return null;
        }
        return {
            name: name.slice(annotationPrefix.length),
            rangeString,
            query,
        };
    };
    const { code: codeWithoutComments, annotations } = await extractAnnotations(
    // TODO hack until we fix out of range annotations in lighter
    code + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n", lang, extractor);
    return { code: codeWithoutComments.trim(), annotations };
}
async function extractLineAnnotations(code) {
    const annotations = code.split("\n").map((line, i) => ({
        name: "Line",
        query: `${i + 1}`,
        ranges: [
            {
                fromLineNumber: i + 1,
                toLineNumber: i + 1,
            },
        ],
    }));
    return { code, annotations };
}
// async function extractTwoSlashAnnotations(code: string, lang: string) {
//   if (lang !== "ts") {
//     return { code, annotations: [] }
//   }
//   const settings = {}
//   const result = twoslasher(code, lang, settings)
//   const { staticQuickInfos, queries } = result
//   console.log(queries)
//   const output = {
//     code: result.code,
//     annotations: [
//       ...staticQuickInfos.map(infoToAnnotation),
//       ...queries.map(queryToAnnotation),
//     ],
//   }
//   return output
// }
// function infoToAnnotation({
//   text,
//   line,
//   character,
//   length,
// }: TwoSlashReturn["staticQuickInfos"][0]): Annotation {
//   return {
//     name: "TypeInfo",
//     query: text,
//     ranges: [
//       {
//         lineNumber: line + 1,
//         fromColumn: character + 1,
//         toColumn: character + length,
//       },
//     ],
//   }
// }
// function queryToAnnotation({
//   text,
//   line,
//   offset,
// }: TwoSlashReturn["queries"][0]): Annotation {
//   return {
//     name: "TypeQuery",
//     query: offset + " " + text,
//     ranges: [
//       {
//         fromLineNumber: line,
//         toLineNumber: line,
//       },
//     ],
//   }
// }
async function readFile(externalCodePath, mdxFilePath, range) {
    const annotationContent = "from " + mdxFilePath + " " + (range || "");
    let fs, path;
    try {
        fs = (await import("fs")).default;
        path = (await import("path")).default;
        if (!fs || !fs.readFileSync || !path || !path.resolve) {
            throw new Error("fs or path not found");
        }
    }
    catch (e) {
        e.message = `Code Hike couldn't resolve this annotation:
${annotationContent}
Looks like node "fs" and "path" modules are not available.`;
        throw e;
    }
    // if we don't know the path of the mdx file:
    if (mdxFilePath == null) {
        throw new Error(`Code Hike couldn't resolve this annotation:
  ${annotationContent}
  Someone is calling the mdx compile function without setting the path.
  Open an issue on CodeHike's repo for help.`);
    }
    const dir = path.dirname(mdxFilePath);
    const absoluteCodepath = path.resolve(dir, externalCodePath);
    let content;
    try {
        content = fs.readFileSync(absoluteCodepath, "utf8");
    }
    catch (e) {
        e.message = `Code Hike couldn't resolve this annotation:
${annotationContent}
${absoluteCodepath} doesn't exist.`;
        throw e;
    }
    if (range) {
        const [start, end] = range.split(":");
        const startLine = parseInt(start);
        const endLine = parseInt(end);
        if (isNaN(startLine) || isNaN(endLine)) {
            throw new Error(`Code Hike couldn't resolve this annotation:
${annotationContent}
The range is not valid. Should be something like:
 ${externalCodePath} 2:5`);
        }
        const lines = content.split("\n");
        content = lines.slice(startLine - 1, endLine).join("\n");
    }
    return content;
}
